# coding=utf-8
# Copyright 2020 The Gin-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Init file for Gin."""
from gin.config import add_config_file_search_path
from gin.config import bind_parameter
from gin.config import clear_config
from gin.config import config_is_locked
from gin.config import config_scope
from gin.config import config_str
from gin.config import configurable
from gin.config import constant
from gin.config import constants_from_enum
from gin.config import current_scope
from gin.config import current_scope_str
from gin.config import enter_interactive_mode
from gin.config import exit_interactive_mode
from gin.config import external_configurable
from gin.config import finalize
from gin.config import operative_config_str
from gin.config import parse_config
from gin.config import parse_config_file
from gin.config import parse_config_files_and_bindings
from gin.config import query_parameter
from gin.config import register
from gin.config import REQUIRED
from gin.config import unlock_config
import gin.resource_reader
